#include <atomic>
#include <cstring>
#include <ctime>
#include <functional>
#include <memory>
#include <signal.h>
#include <stdexcept>

#ifndef TYPES_H
#define TYPES_H
template <typename... Args> class Timer {
public:
  using TimerCallback = std::function<void(Args...)>;

  enum class TimerType { SINGLE_SHOT, PERIODIC };

private:
  struct CallbackContext {
    Timer *timer;
    TimerCallback callback;
    std::tuple<Args...> args;

    CallbackContext(Timer *t, TimerCallback cb, Args... arguments)
        : timer(t), callback(cb), args(std::make_tuple(arguments...)) {}
  };

  timer_t timerId;
  struct sigevent sev;
  struct itimerspec its;
  std::unique_ptr<CallbackContext> context;
  std::atomic<bool> isActive{false};
  std::atomic<bool> isCreated{false};
  TimerType timerType;

  static void timerExpired(union sigval sv) {
    CallbackContext *ctx = static_cast<CallbackContext *>(sv.sival_ptr);
    if (ctx && ctx->timer && ctx->timer->isActive.load()) {
      // Unpack tuple and call callback with arguments
      std::apply(ctx->callback, ctx->args);

      if (ctx->timer->timerType == TimerType::SINGLE_SHOT) {
        ctx->timer->isActive.store(false);
      }
    }
  }

public:
  Timer() {
    std::memset(&sev, 0, sizeof(sev));
    std::memset(&its, 0, sizeof(its));

    sev.sigev_notify = SIGEV_THREAD;
    sev.sigev_notify_function = timerExpired;
  }

  ~Timer() {
    stop();
    if (isCreated.load()) {
      timer_delete(timerId);
    }
  }

  /**
   * Arm timer with callback and arguments
   */
  void arm(int delayMs, int intervalMs, TimerCallback cb, Args... args,
           TimerType type = TimerType::SINGLE_SHOT) {
    if (!cb) {
      throw std::invalid_argument("Callback function cannot be null");
    }

    stop();

    // Create context with callback and arguments
    context = std::make_unique<CallbackContext>(this, cb, args...);
    timerType = type;

    if (!isCreated.load()) {
      sev.sigev_value.sival_ptr = context.get();
      if (timer_create(CLOCK_MONOTONIC, &sev, &timerId) == -1) {
        throw std::runtime_error("Failed to create POSIX timer");
      }
      isCreated.store(true);
    }

    // Configure timer
    its.it_value.tv_sec = delayMs / 1000;
    its.it_value.tv_nsec = (delayMs % 1000) * 1000000;

    if (type == TimerType::PERIODIC && intervalMs > 0) {
      its.it_interval.tv_sec = intervalMs / 1000;
      its.it_interval.tv_nsec = (intervalMs % 1000) * 1000000;
    } else {
      its.it_interval.tv_sec = 0;
      its.it_interval.tv_nsec = 0;
    }

    if (timer_settime(timerId, 0, &its, nullptr) == -1) {
      throw std::runtime_error("Failed to set timer");
    }

    isActive.store(true);
  }

  /**
   * Reset the timer - restarts with the same configuration
   */
  void reset() {
    if (!isCreated.load() || !isActive.load()) {
      throw std::runtime_error("Timer not armed or not active");
    }

    // Re-arm the timer with the same settings stored in 'its'
    if (timer_settime(timerId, 0, &its, nullptr) == -1) {
      throw std::runtime_error("Failed to reset timer");
    }

    // For single-shot timers that were previously expired, reactivate them
    if (timerType == TimerType::SINGLE_SHOT) {
      isActive.store(true);
    }
  }

  void stop() {
    if (isCreated.load() && isActive.load()) {
      struct itimerspec disarm = {{0, 0}, {0, 0}};
      timer_settime(timerId, 0, &disarm, nullptr);
      isActive.store(false);
    }
  }

  bool active() const { return isActive.load(); }
};

#endif